--
-- Advanced Wood Crusher
--
-- @author Queen of Modding
-- @version 1.0.0.0
-- @date 02/11/2020

AdvancedWoodCrusher = {}
AdvancedWoodCrusher.advancedStatisticsPrefix = "AdvancedWoodCrusher"
AdvancedWoodCrusher.advancedStatistics = {{"CrushedLitres", 2}, {"CrushedTrunks", 0}}

function AdvancedWoodCrusher.prerequisitesPresent(specializations)
    return SpecializationUtil.hasSpecialization(TurnOnVehicle, specializations) and SpecializationUtil.hasSpecialization(FillUnit, specializations) and SpecializationUtil.hasSpecialization(PowerConsumer, specializations) and
        not SpecializationUtil.hasSpecialization(WoodCrusher, specializations)
end

function AdvancedWoodCrusher.initSpecialization()
    if g_advancedStatsManager ~= nil then
        AdvancedWoodCrusher.advancedStatistics = g_advancedStatsManager:registerStatistics(AdvancedWoodCrusher.advancedStatisticsPrefix, AdvancedWoodCrusher.advancedStatistics)
    end
end

function AdvancedWoodCrusher.registerEvents(vehicleType)
    SpecializationUtil.registerEvent(vehicleType, "onCrushedSplitShape")
end

function AdvancedWoodCrusher.registerFunctions(vehicleType)
    SpecializationUtil.registerFunction(vehicleType, "crushSplitShape", AdvancedWoodCrusher.crushSplitShape)
    SpecializationUtil.registerFunction(vehicleType, "updateMainDrumRefNodeY", AdvancedWoodCrusher.updateMainDrumRefNodeY)
    SpecializationUtil.registerFunction(vehicleType, "AdvancedWoodCrusherSplitShapeCallback", AdvancedWoodCrusher.AdvancedWoodCrusherSplitShapeCallback)
    SpecializationUtil.registerFunction(vehicleType, "AdvancedWoodCrusherMoveTriggerCallback", AdvancedWoodCrusher.AdvancedWoodCrusherMoveTriggerCallback)
end

function AdvancedWoodCrusher.registerOverwrittenFunctions(vehicleType)
    SpecializationUtil.registerOverwrittenFunction(vehicleType, "getDirtMultiplier", AdvancedWoodCrusher.getDirtMultiplier)
    SpecializationUtil.registerOverwrittenFunction(vehicleType, "getWearMultiplier", AdvancedWoodCrusher.getWearMultiplier)
    SpecializationUtil.registerOverwrittenFunction(vehicleType, "getCanBeTurnedOn", AdvancedWoodCrusher.getCanBeTurnedOn)
    SpecializationUtil.registerOverwrittenFunction(vehicleType, "getConsumingLoad", AdvancedWoodCrusher.getConsumingLoad)
end

function AdvancedWoodCrusher.registerEventListeners(vehicleType)
    SpecializationUtil.registerEventListener(vehicleType, "onLoad", AdvancedWoodCrusher)
    SpecializationUtil.registerEventListener(vehicleType, "onDelete", AdvancedWoodCrusher)
    SpecializationUtil.registerEventListener(vehicleType, "onReadUpdateStream", AdvancedWoodCrusher)
    SpecializationUtil.registerEventListener(vehicleType, "onWriteUpdateStream", AdvancedWoodCrusher)
    SpecializationUtil.registerEventListener(vehicleType, "onUpdate", AdvancedWoodCrusher)
    SpecializationUtil.registerEventListener(vehicleType, "onUpdateTick", AdvancedWoodCrusher)
    SpecializationUtil.registerEventListener(vehicleType, "onTurnedOn", AdvancedWoodCrusher)
    SpecializationUtil.registerEventListener(vehicleType, "onTurnedOff", AdvancedWoodCrusher)
    SpecializationUtil.registerEventListener(vehicleType, "onCrushedSplitShape", AdvancedWoodCrusher)
end

function AdvancedWoodCrusher:onLoad(savegame)
    self.spec_advancedWoodCrusher = {} -- only needed for mods
    local spec = self.spec_advancedWoodCrusher

    local xmlRoot = getXMLRootName(self.xmlFile)

    XMLUtil.checkDeprecatedXMLElements(self.xmlFile, self.configFileName, xmlRoot .. ".woodCrusher.moveTrigger(0)#index", xmlRoot .. ".woodCrusher.moveTriggers.trigger#node") --FS17 to FS19
    XMLUtil.checkDeprecatedXMLElements(self.xmlFile, self.configFileName, xmlRoot .. ".woodCrusher.moveCollision(0)#index", xmlRoot .. ".woodCrusher.moveCollisions.collision#node") --FS17 to FS19
    XMLUtil.checkDeprecatedXMLElements(self.xmlFile, self.configFileName, xmlRoot .. ".woodCrusher.emitterShape(0)", xmlRoot .. ".woodCrusher.crushEffects with effectClass 'ParticleEffect'") --FS17 to FS19
    XMLUtil.checkDeprecatedXMLElements(self.xmlFile, self.configFileName, xmlRoot .. ".AdvancedWoodCrusherStartSound", xmlRoot .. ".woodCrusher.sounds.start") --FS17 to FS19
    XMLUtil.checkDeprecatedXMLElements(self.xmlFile, self.configFileName, xmlRoot .. ".AdvancedWoodCrusherIdleSound", xmlRoot .. ".woodCrusher.sounds.idle") --FS17 to FS19
    XMLUtil.checkDeprecatedXMLElements(self.xmlFile, self.configFileName, xmlRoot .. ".AdvancedWoodCrusherWorkSound", xmlRoot .. ".woodCrusher.sounds.work") --FS17 to FS19
    XMLUtil.checkDeprecatedXMLElements(self.xmlFile, self.configFileName, xmlRoot .. ".AdvancedWoodCrusherStopSound", xmlRoot .. ".woodCrusher.sounds.stop") --FS17 to FS19
    XMLUtil.checkDeprecatedXMLElements(self.xmlFile, self.configFileName, xmlRoot .. ".turnedOnRotationNodes.turnedOnRotationNode#type", xmlRoot .. ".woodCrusher.animationNodes.animationNode", "AdvancedWoodCrusher") --FS17 to FS19
    XMLUtil.checkDeprecatedXMLElements(self.xmlFile, self.configFileName, xmlRoot .. ".turnedOnScrollers.turnedOnScroller", xmlRoot .. ".woodCrusher.animationNodes.animationNode") --FS17 to FS19

    local baseKey = xmlRoot .. ".woodCrusher"
    spec.cutNode = I3DUtil.indexToObject(self.components, getXMLString(self.xmlFile, baseKey .. "#cutNode"), self.i3dMappings)
    spec.mainDrumRefNode = I3DUtil.indexToObject(self.components, getXMLString(self.xmlFile, baseKey .. "#mainDrumRefNode"), self.i3dMappings)
    spec.mainDrumRefNodeY = 0
    if spec.mainDrumRefNode ~= nil then
        _, spec.mainDrumRefNodeY, _ = getTranslation(spec.mainDrumRefNode)
    end

    spec.moveTriggers = {}
    local i = 0
    while true do
        local key = string.format("%s.moveTriggers.trigger(%d)", baseKey, i)
        if not hasXMLProperty(self.xmlFile, key) then
            break
        end

        local node = I3DUtil.indexToObject(self.components, getXMLString(self.xmlFile, key .. "#node"), self.i3dMappings)
        if node ~= nil then
            table.insert(spec.moveTriggers, node)
        end
        i = i + 1
    end

    spec.moveColNodes = {}
    i = 0
    while true do
        local key = string.format("%s.moveCollisions.collision(%d)", baseKey, i)
        if not hasXMLProperty(self.xmlFile, key) then
            break
        end

        local node = I3DUtil.indexToObject(self.components, getXMLString(self.xmlFile, key .. "#node"), self.i3dMappings)
        if node ~= nil then
            table.insert(spec.moveColNodes, node)
        end
        i = i + 1
    end

    spec.moveVelocityZ = Utils.getNoNil(getXMLFloat(self.xmlFile, baseKey .. "#moveVelocityZ"), 0.8) -- m/s
    spec.downForceNode = I3DUtil.indexToObject(self.components, getXMLString(self.xmlFile, baseKey .. "#downForceNode"), self.i3dMappings)
    spec.downForce = Utils.getNoNil(getXMLFloat(self.xmlFile, baseKey .. "#downForce"), 2)
    spec.cutSizeY = Utils.getNoNil(getXMLFloat(self.xmlFile, baseKey .. "#cutSizeY"), 1)
    spec.cutSizeZ = Utils.getNoNil(getXMLFloat(self.xmlFile, baseKey .. "#cutSizeZ"), 1)

    spec.moveTriggerNodes = {}
    if self.isServer and spec.moveTriggers ~= nil then
        for _, node in pairs(spec.moveTriggers) do
            addTrigger(node, "AdvancedWoodCrusherMoveTriggerCallback", self)
        end
    end

    spec.crushNodes = {}
    spec.crushingTime = 0
    spec.turnOnAutomatically = Utils.getNoNil(getXMLBool(self.xmlFile, baseKey .. "#automaticallyTurnOn"), false)

    if self.isClient then
        spec.crushEffects = g_effectManager:loadEffect(self.xmlFile, baseKey .. ".crushEffects", self.components, self, self.i3dMappings)

        spec.animationNodes = g_animationManager:loadAnimations(self.xmlFile, baseKey .. ".animationNodes", self.components, self, self.i3dMappings)

        spec.isWorkSamplePlaying = false
        spec.samples = {}
        spec.samples.start = g_soundManager:loadSampleFromXML(self.xmlFile, baseKey .. ".sounds", "start", self.baseDirectory, self.components, 1, AudioGroup.VEHICLE, self.i3dMappings, self)
        spec.samples.stop = g_soundManager:loadSampleFromXML(self.xmlFile, baseKey .. ".sounds", "stop", self.baseDirectory, self.components, 1, AudioGroup.VEHICLE, self.i3dMappings, self)
        spec.samples.work = g_soundManager:loadSampleFromXML(self.xmlFile, baseKey .. ".sounds", "work", self.baseDirectory, self.components, 0, AudioGroup.VEHICLE, self.i3dMappings, self)
        spec.samples.idle = g_soundManager:loadSampleFromXML(self.xmlFile, baseKey .. ".sounds", "idle", self.baseDirectory, self.components, 0, AudioGroup.VEHICLE, self.i3dMappings, self)
    end

    local moveColDisableCollisionPairs = Utils.getNoNil(getXMLBool(self.xmlFile, "vehicle.woodCrusher#moveColDisableCollisionPairs"), true)
    if moveColDisableCollisionPairs then
        for _, component in pairs(self.components) do
            for _, node in pairs(spec.moveColNodes) do
                setPairCollision(component.node, node, false)
            end
        end
    end

    spec.fillUnitIndex = Utils.getNoNil(getXMLInt(self.xmlFile, "vehicle.woodCrusher#fillUnitIndex"), 1)
end

function AdvancedWoodCrusher:onDelete()
    local spec = self.spec_advancedWoodCrusher
    if self.isServer and spec.moveTriggers ~= nil then
        for _, node in pairs(spec.moveTriggers) do
            removeTrigger(node)
        end
    end

    if self.isClient then
        g_effectManager:deleteEffects(spec.crushEffects)
        g_soundManager:deleteSamples(spec.samples)
        g_animationManager:deleteAnimations(spec.animationNodes)
    end
end

function AdvancedWoodCrusher:onReadUpdateStream(streamId, timestamp, connection)
    if connection:getIsServer() then
        local spec = self.spec_advancedWoodCrusher
        if streamReadBool(streamId) then
            spec.crushingTime = 1000
        else
            spec.crushingTime = 0
        end
        spec.mainDrumRefNodeY = streamReadFloat32(streamId)
    end
end

function AdvancedWoodCrusher:onWriteUpdateStream(streamId, connection, dirtyMask)
    if not connection:getIsServer() then
        local spec = self.spec_advancedWoodCrusher
        streamWriteBool(streamId, spec.crushingTime > 0)
        streamWriteFloat32(streamId, spec.mainDrumRefNodeY)
    end
end

function AdvancedWoodCrusher:onUpdate(dt, isActiveForInput, isActiveForInputIgnoreSelection, isSelected)
    local spec = self.spec_advancedWoodCrusher

    if self.isServer then
        if self:getIsTurnedOn() then
            for node in pairs(spec.crushNodes) do
                self:crushSplitShape(node)
                spec.crushNodes[node] = nil
                spec.moveTriggerNodes[node] = nil
            end
        end

        local maxTreeSizeY = 0
        for id in pairs(spec.moveTriggerNodes) do
            if not entityExists(id) then
                spec.moveTriggerNodes[id] = nil
            else
                if spec.downForceNode ~= nil then
                    local x1, y1, z1 = getWorldTranslation(spec.downForceNode)
                    local nx, ny, nz = localDirectionToWorld(spec.downForceNode, 1, 0, 0)
                    local yx, yy, yz = localDirectionToWorld(spec.downForceNode, 0, 1, 0)

                    local minY, maxY, minZ, maxZ = testSplitShape(id, x1, y1, z1, nx, ny, nz, yx, yy, yz, spec.cutSizeY, spec.cutSizeZ)
                    if minY ~= nil then
                        local cx, cy, cz = localToWorld(spec.downForceNode, 0, (minY + maxY) * 0.5, (minZ + maxZ) * 0.5)
                        local downX, downY, downZ = localDirectionToWorld(spec.downForceNode, 0, -spec.downForce, 0)
                        addForce(id, downX, downY, downZ, cx, cy, cz, false)
                        maxTreeSizeY = math.max(maxTreeSizeY, maxY)
                    end
                end
            end
        end

        if spec.mainDrumRefNode ~= nil then
            spec.mainDrumRefNodeY = 0
            if maxTreeSizeY > 0 then
                local x, y, z = localToWorld(spec.downForceNode, 0, maxTreeSizeY * 1.15, 0)
                _, spec.mainDrumRefNodeY, _ = worldToLocal(getParent(spec.mainDrumRefNode), x, y, z)
            end
        end

        if next(spec.moveTriggerNodes) ~= nil or spec.crushingTime > 0 then
            self:raiseActive()
        end
    end
    self:updateMainDrumRefNodeY(dt)
end

function AdvancedWoodCrusher:updateMainDrumRefNodeY(dt)
    local spec = self.spec_advancedWoodCrusher
    if spec.mainDrumRefNode ~= nil then
        local x, y, z = getTranslation(spec.mainDrumRefNode)
        if spec.mainDrumRefNodeY > y then
            y = math.min(y + 0.0008 * dt, spec.mainDrumRefNodeY) -- up
        else
            y = math.max(y - 0.0005 * dt, spec.mainDrumRefNodeY) -- down
        end
        setTranslation(spec.mainDrumRefNode, x, y, z)
    end
end

function AdvancedWoodCrusher:onUpdateTick(dt, isActiveForInput, isActiveForInputIgnoreSelection, isSelected)
    local spec = self.spec_advancedWoodCrusher

    if self:getIsTurnedOn() then
        if self.isServer then
            if spec.cutNode ~= nil and next(spec.moveTriggerNodes) ~= nil then
                local x, y, z = getWorldTranslation(spec.cutNode)
                local nx, ny, nz = localDirectionToWorld(spec.cutNode, 1, 0, 0)
                local yx, yy, yz = localDirectionToWorld(spec.cutNode, 0, 1, 0)
                for id in pairs(spec.moveTriggerNodes) do
                    local lenBelow, lenAbove = getSplitShapePlaneExtents(id, x, y, z, nx, ny, nz)
                    if lenAbove ~= nil and lenBelow ~= nil then
                        if lenBelow <= 0.4 then
                            spec.moveTriggerNodes[id] = nil
                            self:crushSplitShape(id)
                        elseif lenAbove >= 0.2 then
                            local minY = splitShape(id, x, y, z, nx, ny, nz, yx, yy, yz, spec.cutSizeY, spec.cutSizeZ, "AdvancedWoodCrusherSplitShapeCallback", self)
                            if minY ~= nil then
                                spec.moveTriggerNodes[id] = nil
                            end
                        end
                    end
                end
            end
        end
    end

    if spec.crushingTime > 0 then
        spec.crushingTime = math.max(spec.crushingTime - dt, 0)
    end

    local isCrushing = spec.crushingTime > 0

    if self.isClient then
        if isCrushing then
            g_effectManager:setFillType(spec.crushEffects, FillType.WOODCHIPS)
            g_effectManager:startEffects(spec.crushEffects)
        else
            g_effectManager:stopEffects(spec.crushEffects)
        end

        if self:getIsTurnedOn() and isCrushing then
            if not spec.isWorkSamplePlaying then
                g_soundManager:playSample(spec.samples.work)
                spec.isWorkSamplePlaying = true
            end
        else
            if spec.isWorkSamplePlaying then
                g_soundManager:stopSample(spec.samples.work)
                spec.isWorkSamplePlaying = false
            end
        end
    end

    -- turn on/off automatically if tree node is in trigger
    if self.isServer then
        if g_currentMission.missionInfo.automaticMotorStartEnabled then
            if spec.turnOnAutomatically and self.setIsTurnedOn ~= nil then
                if next(spec.moveTriggerNodes) ~= nil then
                    if self.getIsMotorStarted ~= nil then
                        if not self:getIsMotorStarted() then
                            self:startMotor()
                        end
                    else
                        if self.attacherVehicle ~= nil then
                            if self.attacherVehicle.getIsMotorStarted ~= nil then
                                if not self.attacherVehicle:getIsMotorStarted() then
                                    self.attacherVehicle:startMotor()
                                end
                            end
                        end
                    end

                    if not self.isControlled and not self:getIsTurnedOn() and self:getCanBeTurnedOn() then
                        self:setIsTurnedOn(true)
                    end
                    spec.turnOffTimer = 3000
                else
                    if self:getIsTurnedOn() then
                        if spec.turnOffTimer == nil then
                            spec.turnOffTimer = 3000
                        end
                        spec.turnOffTimer = spec.turnOffTimer - dt

                        if spec.turnOffTimer < 0 then
                            local rootAttacherVehicle = self:getRootVehicle()

                            if not rootAttacherVehicle.isControlled then
                                if self.getIsMotorStarted ~= nil then
                                    if self:getIsMotorStarted() then
                                        self:stopMotor()
                                    end
                                end

                                self:setIsTurnedOn(false)
                            end
                        end
                    end
                end
            end
        end
    end
end

function AdvancedWoodCrusher:onTurnedOn()
    local spec = self.spec_advancedWoodCrusher

    if self.isServer and spec.moveColNodes ~= nil then
        for _, node in pairs(spec.moveColNodes) do
            setFrictionVelocity(node, spec.moveVelocityZ)
        end
    end

    if self.isClient then
        g_soundManager:stopSamples(spec.samples)
        spec.isWorkSamplePlaying = false
        g_soundManager:playSample(spec.samples.start)
        g_soundManager:playSample(spec.samples.idle, 0, spec.samples.start)

        if self.isClient then
            g_animationManager:startAnimations(spec.animationNodes)
        end
    end
end

function AdvancedWoodCrusher:onTurnedOff()
    local spec = self.spec_advancedWoodCrusher

    if self.isServer then
        for node in pairs(spec.crushNodes) do
            self:crushSplitShape(node)
            spec.crushNodes[node] = nil
        end
        if spec.moveColNodes ~= nil then
            for _, node in pairs(spec.moveColNodes) do
                setFrictionVelocity(node, 0.0)
            end
        end
    end

    if self.isClient then
        g_effectManager:stopEffects(spec.crushEffects)
        g_soundManager:stopSamples(spec.samples)
        g_soundManager:playSample(spec.samples.stop)
        spec.isWorkSamplePlaying = false

        if self.isClient then
            g_animationManager:stopAnimations(spec.animationNodes)
        end
    end
end

function AdvancedWoodCrusher:getCanBeTurnedOn(superFunc)
    local spec = self.spec_advancedWoodCrusher
    if spec.turnOnAutomatically then
        return false
    end

    return superFunc(self)
end

function AdvancedWoodCrusher:getDirtMultiplier(superFunc)
    local multiplier = superFunc(self)

    local spec = self.spec_advancedWoodCrusher
    if spec.crushingTime > 0 then
        multiplier = multiplier + self:getWorkDirtMultiplier()
    end

    return multiplier
end

function AdvancedWoodCrusher:getWearMultiplier(superFunc)
    local multiplier = superFunc(self)

    local spec = self.spec_advancedWoodCrusher
    if spec.crushingTime > 0 then
        multiplier = multiplier + self:getWorkWearMultiplier()
    end

    return multiplier
end

function AdvancedWoodCrusher:onCrushedSplitShape(splitType, volume)
    local spec = self.spec_advancedWoodCrusher
    self:addFillUnitFillLevel(self:getOwnerFarmId(), spec.fillUnitIndex, volume * 1000 * splitType.woodChipsPerLiter, FillType.WOODCHIPS, ToolType.UNDEFINED)
    if g_advancedStatsManager ~= nil then
        g_advancedStatsManager.updateStatistic(self, AdvancedWoodCrusher.advancedStatistics["CrushedTrunks"], 1)
        g_advancedStatsManager.updateStatistic(self, AdvancedWoodCrusher.advancedStatistics["CrushedLitres"], volume * 1000 * splitType.woodChipsPerLiter)
    end
end

function AdvancedWoodCrusher:crushSplitShape(shape)
    local splitType = g_splitTypeManager:getSplitTypeByIndex(getSplitType(shape))
    if splitType ~= nil and splitType.woodChipsPerLiter > 0 then
        local spec = self.spec_advancedWoodCrusher
        local volume = getVolume(shape)
        delete(shape)
        spec.crushingTime = 1000
        SpecializationUtil.raiseEvent(self, "onCrushedSplitShape", splitType, volume)
    end
end

function AdvancedWoodCrusher:AdvancedWoodCrusherSplitShapeCallback(shape, isBelow, isAbove, minY, maxY, minZ, maxZ)
    local spec = self.spec_advancedWoodCrusher
    if not isBelow then
        spec.crushNodes[shape] = shape
    end
end

function AdvancedWoodCrusher:AdvancedWoodCrusherMoveTriggerCallback(triggerId, otherActorId, onEnter, onLeave, onStay, otherShapeId)
    local vehicle = g_currentMission.nodeToObject[otherActorId]
    if vehicle == nil and getRigidBodyType(otherActorId) == "Dynamic" then
        local splitType = g_splitTypeManager:getSplitTypeByIndex(getSplitType(otherActorId))
        if splitType ~= nil and splitType.woodChipsPerLiter > 0 then
            local spec = self.spec_advancedWoodCrusher
            if onEnter then
                spec.moveTriggerNodes[otherActorId] = Utils.getNoNil(spec.moveTriggerNodes[otherActorId], 0) + 1
                self:raiseActive()
            elseif onLeave then
                local c = spec.moveTriggerNodes[otherActorId]
                if c ~= nil then
                    c = c - 1
                    if c == 0 then
                        spec.moveTriggerNodes[otherActorId] = nil
                    else
                        spec.moveTriggerNodes[otherActorId] = c
                    end
                end
            end
        end
    end
end

function AdvancedWoodCrusher:getConsumingLoad(superFunc)
    local value, count = superFunc(self)

    local spec = self.spec_advancedWoodCrusher

    local loadPercentage = 0.001
    if spec.crushingTime > 0 then
        loadPercentage = 1
    end

    return value + loadPercentage, count + 1
end
